function W=finvariant(pos,T,fs,M,nPbasis,nPdesign,theta_test,y_test,invar,beta,show)

% W = finvariant(pos,T,fs,M,nPbasis,nPdesign,theta_test,y_test,invar,beta,show)
% demonstrates the frequency invariant basis for a linear array according to
% (Parra, JASA 119 (6), pp. 3839-3847, June 2006). Array positions are given
% in vector pos measured in meters (speed of sound in air is assumed). T is
% the number of filter coefficients in the time domain, and fs the sampling
% rate.  M is the number of Legendre bases (or 'virtual sensors'). To use
% 'naive' non-starable basis with one virtual sensor per anlge set
% M==0. nPbasis is the power of the noise when computing the pseudo-inverse
% that determines the basis transform. nPdesign is the regularization
% constant when computing the pinverse in the LS optimal filter
% design. theta_test y_test and are the desired angle and magnitude of the
% desired response of the LS filter design. invar is a binary variable
% indicating that the same coefficients are to be choosen for all
% frequencies (frequency invariance). beta specifies an angle by which to
% shift the beam. If show is set some diagnostic information will be
% plotted. The function returns in W the time domain filter coefficients for
% each sensor.
% 
%  
% (c) Lucas C. Parra, August 12 2005
%                     Feb 26 2006, changed to equidistant sampling in theta

Dx=length(pos);   % number of microphones
TT=T/2+1;         % number of freq. bins

equi_theta=1; % equidistant sampling in theta vs cos(theta) for computing
              % the basis transform (sampling for the rotation matrix is
              % determined separatelly in legshift() )

% define sampling for Legendre basis
if equi_theta
  theta=[-89:89]; 
  x = sin(theta/180*pi);
else
  x=[-1:0.01:1];          % equidist. discret. of Lagendre basis
  theta=asin(x)/pi*180; % corresponding angles
end

% broadsize corresponds to theta=0. This is not the conventional definition
% for the Legendre Polynomials. Some graphs may vary to match with the
% convetional definition of theta=pi for broadside.
		
% compute array response and values of freq. bins
G=arrayresp(theta,pos,fs,T,T/2);
freq=(0:T/2)/T*fs;                                                                                
% compute basis transform that generates freq invariant responses 
I = nPbasis*repmat(shiftdim(eye(Dx),-1),[TT 1 1]);
Ginv = ip(hermite(G),mypinv(ip(G,hermite(G))+I));

% compute the basis transform that maps to a legendre (stearable basis)
if M
  % compute Legendre basis
  P = legdecomp(x,M); 
  % transform basis
  for i=1:Dx, PGinv(:,:,i) = Ginv(:,:,i)*P; end;
else 
  % keep 'naive' basis (one virtual sensor pwer angle)
  PGinv = Ginv; 
end

% now solve the optimal LS filter design in the new bases

% target response
R = permute(repmat(y_test,[T/2+1 1]),[1 3 2]);

% array response to the test angles
G_test = arrayresp(theta_test,pos,fs,T,T/2);

% MM is rank-deficient! rank == Dx, dim=Dth*Dth
% to invert it we will add this regularizer
Dth = length(theta_test);
Itest = nPdesign*repmat(shiftdim(eye(Dth),-1),[TT 1 1]);

% compute the optimal LS coefficients in each each frequency
MM = ip(PGinv,G_test);
Minv = ip(mypinv(ip(hermite(MM),MM)+Itest),hermite(MM));
A = ip(R,Minv);

% takes the average of the separate solutions for each frequency and leave
% non-invertable freq with high condition number out.
if invar 

  % determine freq for which condition number not too large
  E=ip(G,hermite(G))+I;
  for i=1:size(E,1), f(i) = db(cond(squeeze(E(i,:,:)))); end;
  cond_thresh=min(max(f)-20,150); % ad-hock threshold

  % take the mean of the useful frequencies
  A = repmat(mean(A(f<cond_thresh,:,:)),[TT 1 1]);

  if show, % diagnostic
    subplot(2,2,3);
    plot(freq,f); title('condition number of (G^H G + \sigma^2 I)'); ...
	xlabel('freq (Hz)')
  end

end

% steering
if beta

  % Linear tranform that approximatly shifts the legendre basis
  D = legshift(beta/180*pi,M);
  
  % shift beam by appling to existing coefficients
  D = repmat(shiftdim(D,-1),[TT 1 1]);
  A = ip(A,D);

end

% compute the resuling respones and time domain filter coefficients
H = ip(A,PGinv);
Rest = ip(H,G);
W = real(ifft(cat(1,H,conj(H(TT-1:-1:2,:,:)))));

% show the result
subplot(2,2,1)
plot((1:T)/fs*1000,W(:,:)); axis tight;
title('Filter Coefficients');
xlabel('tap delay (ms)')

subplot(2,2,2)
colormap('hot')
polarresponce(W,[],pos',fs);
title('Array Response')
caxis([-50 0]+max(db(eps+abs(Rest(:)).^2)))
h=colorbar('vert'); set(get(h,'title'),'string','dB')


